@@ -1,11 +1,39 @@
 /*
- * Copyright (c) 2007, Erik Tews, Andrei Pychkine and Ralf-Philipp Weinmann.
+ *  Copyright (c) 2007, 2008, 2009 Erik Tews, Andrei Pychkine and Ralf-Philipp Weinmann.
  *
+ *  This program is free software; you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License as published by
+ *  the Free Software Foundation; either version 2 of the License, or
+ *  (at your option) any later version.
+ *
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with this program; if not, write to the Free Software
+ *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ *
+ *
+ *  In addition, as a special exception, the copyright holders give
+ *  permission to link the code of portions of this program with the
+ *  OpenSSL library under certain conditions as described in each
+ *  individual source file, and distribute linked combinations
+ *  including the two.
+ *  You must obey the GNU General Public License in all respects
+ *  for all of the code used other than OpenSSL. *  If you modify
+ *  file(s) with this exception, you may extend this exception to your
+ *  version of the file(s), but you are not obligated to do so. *  If you
+ *  do not wish to do so, delete this exception statement from your
+ *  version. *  If you delete this exception statement from all source
+ *  files in the program, then also delete it here.
  */
 
 #include <string.h>
 #include <stdio.h>
 #include <stdlib.h>
+#include <pthread.h>
 #include "pcap.h"
 #include "aircrack-ptw-lib.h"
 #include "aircrack-ng.h"
@@ -83,6 +111,10 @@ static const double eval[] = {
 0.00495094196451801,
 0.0048983441590402};
 
+int tried, max_tries;
+int depth[KEYHSBYTES];
+PTW_tableentry keytable[KEYHSBYTES][n];
+
 // For sorting
 static int compare(const void * ina, const void * inb) {
         PTW_tableentry * a = (PTW_tableentry * )ina;
@@ -113,12 +145,17 @@ static int comparedoublesorthelper(const void * ina, const void * inb) {
 // RC4 key setup
 static void rc4init ( uint8_t * key, int keylen, rc4state * state) {
 	int i;
-	int j;
+	unsigned char j;
 	uint8_t tmp;
 	memcpy(state->s, &rc4initial, n);
 	j = 0;
 	for (i = 0; i < n; i++) {
-		j = (j + state->s[i] + key[i % keylen]) % n;
+                /*  this should be:
+                    j = (j + state->s[i] + key[i % keylen]) % n;
+                    but as "j" is declared as unsigned char and n equals 256,
+                    we can "optimize" it
+                */
+		j = (j + state->s[i] + key[i % keylen]);
 		tmp = state->s[i];
 		state->s[i] = state->s[j];
 		state->s[j] = tmp;
@@ -156,21 +193,22 @@ static int comparesorthelper(const void * ina, const void * inb) {
 
 /*
  * Guess the values for sigma_i
+ * ivlen - how long was the iv (is used differently in original klein attack)
  * iv - IV which was used for this packet
  * keystream - keystream recovered
  * result - buffer for the values of sigma_i
  * kb - how many keybytes should be guessed
  */
-static void guesskeybytes(uint8_t * iv, uint8_t * keystream, uint8_t * result, int kb) {
+static void guesskeybytes(int ivlen, uint8_t * iv, uint8_t * keystream, uint8_t * result, int kb) {
         uint8_t state[n];
         uint8_t j = 0;
         uint8_t tmp;
         int i;
-        int jj = IVBYTES;
+        int jj = ivlen;
         uint8_t ii;
         uint8_t s = 0;
         memcpy(state, rc4initial, n);
-        for (i = 0; i < IVBYTES; i++) {
+        for (i = 0; i < ivlen; i++) {
                 j += state[i] + iv[i];
                 tmp = state[i];
                 state[i] = state[j];
@@ -196,10 +234,19 @@ static void guesskeybytes(uint8_t * iv, uint8_t * keystream, uint8_t * result, i
 static int correct(PTW_attackstate * state, uint8_t * key, int keylen) {
 	int i;
         int j;
+        int k;
         uint8_t keybuf[PTW_KSBYTES];
         rc4state rc4state;
 
-        for (i = 0; i < state->sessions_collected; i++) {
+	// We need at least 3 sessions to be somehow certain
+	if (state->sessions_collected < 3) {
+		return 0;
+	}
+
+        tried++;
+
+        k = rand()%(state->sessions_collected-10);
+        for ( i=k; i < k+10; i++) {
                 memcpy(&keybuf[IVBYTES], key, keylen);
                 memcpy(keybuf, state->sessions[i].iv, IVBYTES);
                 rc4init(keybuf, keylen+IVBYTES, &rc4state);
@@ -256,41 +303,54 @@ static void getdrv(PTW_tableentry orgtable[][n], int keylen, double * normal, do
         }
 }
 
-int tried, max_tries;
-int depth[KEYHSBYTES];
-PTW_tableentry keytable[KEYHSBYTES][n];
 /*
  * Guess a single keybyte
  */
-static int doRound(PTW_tableentry sortedtable[][n], int keybyte, int fixat, uint8_t fixvalue, int * searchborders, uint8_t * key, int keylen, PTW_attackstate * state, uint8_t sum, int * strongbytes) {
+static int doRound(PTW_tableentry sortedtable[][n], int keybyte, int fixat, uint8_t fixvalue, int * searchborders, uint8_t * key, int keylen, PTW_attackstate * state, uint8_t sum, int * strongbytes, int * bf, int validchars[][n]) {
 	int i;
 	uint8_t tmp;
 
-	show_wep_stats( keylen -1, 0, keytable, searchborders, depth, tried, max_tries );
-	tried++;
+	if(!opt.is_quiet && keybyte < 4)
+		show_wep_stats( keylen -1, 0, keytable, searchborders, depth, tried );
+	if (keybyte > 0) {
+		if (!validchars[keybyte-1][key[keybyte-1]]) {
+			return 0;
+		}
+	}
 	if (keybyte == keylen) {
 		return correct(state, key, keylen);
+	} else if (bf[keybyte] == 1) {
+		for (i = 0; i < n; i++) {
+			key[keybyte] = i;
+			if (doRound(sortedtable, keybyte+1, fixat, fixvalue, searchborders, key, keylen, state, sum+i%n, strongbytes, bf, validchars)) {
+				return 1;
+			}
+		}
+		return 0;
+        } else if (keybyte == fixat) {
+                key[keybyte] = fixvalue-sum;
+                return doRound(sortedtable, keybyte+1, fixat, fixvalue, searchborders, key, keylen, state, fixvalue, strongbytes, bf, validchars);
 	} else if (strongbytes[keybyte] == 1) {
 		// printf("assuming byte %d to be strong\n", keybyte);
 		tmp = 3 + keybyte;
 		for (i = keybyte-1; i >= 1; i--) {
 			tmp += 3 + key[i] + i;
-			key[keybyte] = 256-tmp;
-			if(doRound(sortedtable, keybyte+1, fixat, fixvalue, searchborders, key, keylen, state, (256-tmp+sum)%256, strongbytes) == 1) {
+			key[keybyte] = n-tmp;
+			if(doRound(sortedtable, keybyte+1, fixat, fixvalue, searchborders, key, keylen, state, (n-tmp+sum)%n, strongbytes, bf, validchars) == 1) {
 				printf("hit with strongbyte for keybyte %d\n", keybyte);
 				return 1;
 			}
 		}
 		return 0;
-	} else if (keybyte == fixat) {
-		key[keybyte] = fixvalue-sum;
-		return doRound(sortedtable, keybyte+1, fixat, fixvalue, searchborders, key, keylen, state, fixvalue, strongbytes);
 	} else {
 		for (i = 0; i < searchborders[keybyte]; i++) {
-			depth[keybyte] = i;
-			key[keybyte] = sortedtable[keybyte][i].b - sum;
-			keytable[keybyte][i].b = key[keybyte];
-			if (doRound(sortedtable, keybyte+1, fixat, fixvalue, searchborders, key, keylen, state, sortedtable[keybyte][i].b, strongbytes) == 1) {
+                    key[keybyte] = sortedtable[keybyte][i].b - sum;
+                    if(!opt.is_quiet)
+                    {
+                        depth[keybyte] = i;
+                        keytable[keybyte][i].b = key[keybyte];
+                    }
+                    if (doRound(sortedtable, keybyte+1, fixat, fixvalue, searchborders, key, keylen, state, sortedtable[keybyte][i].b, strongbytes, bf, validchars)) {
 				return 1;
 			}
 		}
@@ -301,14 +361,15 @@ static int doRound(PTW_tableentry sortedtable[][n], int keybyte, int fixat, uint
 /*
  * Do the actual computation of the key
  */
-static int doComputation(PTW_attackstate * state, uint8_t * key, int keylen, PTW_tableentry table[][n], sorthelper * sh2, int * strongbytes, int keylimit) {
+static int doComputation(PTW_attackstate * state, uint8_t * key, int keylen, PTW_tableentry table[][n], sorthelper * sh2, int * strongbytes, int keylimit, int * bf, int validchars[][n]) {
 	int i,j;
 	int choices[KEYHSBYTES];
 	int prod;
 	int fixat;
 	int fixvalue;
 
-	memcpy(keytable, table, sizeof(PTW_tableentry) * n * keylen);
+        if(!opt.is_quiet)
+            memcpy(keytable, table, sizeof(PTW_tableentry) * n * keylen);
 
 	for (i = 0; i < keylen; i++) {
 		if (strongbytes[i] == 1) {
@@ -321,14 +382,22 @@ static int doComputation(PTW_attackstate * state, uint8_t * key, int keylen, PTW
 	prod = 0;
 	fixat = -1;
 	fixvalue = 0;
-	max_tries = keylimit;
+        max_tries = keylimit;
 
 	while(prod < keylimit) {
-		if (doRound(table, 0, fixat, fixvalue, choices, key, keylen, state, 0, strongbytes) == 1) {
+		if (doRound(table, 0, fixat, fixvalue, choices, key, keylen, state, 0, strongbytes, bf, validchars) == 1) {
 			// printf("hit with %d choices\n", prod);
-			show_wep_stats( keylen -1, 1, keytable, choices, depth, prod, keylimit );
+			if(!opt.is_quiet)
+				show_wep_stats( keylen -1, 1, keytable, choices, depth, tried );
 			return 1;
 		}
+		while( (i < keylen * (n-1)) && ((strongbytes[sh2[i].keybyte] == 1) || (bf[sh2[i].keybyte] == 1) ) ) {
+			i++;
+		}
+		if(i >= (keylen * (n-1)))
+		{
+			break;
+		}
 		choices[sh2[i].keybyte]++;
 		fixat = sh2[i].keybyte;
 		// printf("choices[%d] is now %d\n", sh2[i].keybyte, choices[sh2[i].keybyte]);
@@ -336,88 +405,143 @@ static int doComputation(PTW_attackstate * state, uint8_t * key, int keylen, PTW
 		prod = 1;
 		for (j = 0; j < keylen; j++) {
 			prod *= choices[j];
+			if (bf[j] == 1) {
+				prod *= n;
+			}
 		}
-		tried = prod;
+
+                /*
 		do {
 			i++;
 		} while (strongbytes[sh2[i].keybyte] == 1);
+		*/
+		i++;
 
-		show_wep_stats( keylen -1, 0, keytable, choices, depth, prod, keylimit );
+		if(!opt.is_quiet)
+			show_wep_stats( keylen -1, 0, keytable, choices, depth, tried );
 
 	}
-	show_wep_stats( keylen -1, 1, keytable, choices, depth, prod, keylimit );
-	return 0;
+	if(!opt.is_quiet)
+            show_wep_stats( keylen -1, 1, keytable, choices, depth, tried );
+    return 0;
 }
 
+
 /*
  * Guess which key bytes could be strong and start actual computation of the key
  */
-int PTW_computeKey(PTW_attackstate * state, uint8_t * keybuf, int keylen, int testlimit) {
+int PTW_computeKey(PTW_attackstate * state, uint8_t * keybuf, int keylen, int testlimit, int * bf, int validchars[][n], int attacks) {
 	int strongbytes[KEYHSBYTES];
 	double normal[KEYHSBYTES];
 	double ausreisser[KEYHSBYTES];
 	doublesorthelper helper[KEYHSBYTES];
 	int simple, onestrong, twostrong;
 	int i,j;
-
-	sorthelper (* sh)[n-1] = NULL;
+	uint8_t fullkeybuf[PTW_KSBYTES];
+	uint8_t guessbuf[PTW_KSBYTES];
+	sorthelper(*sh)[n-1];
 	PTW_tableentry (*table)[n] = alloca(sizeof(PTW_tableentry) * n * keylen);
 
+        tried=0;
+	sh = NULL;
+
 	if (table == NULL) {
 		printf("could not allocate memory\n");
 		exit(-1);
 	}
-	memcpy(table, state->table, sizeof(PTW_tableentry) * n * keylen);
 
-	onestrong = (testlimit/10)*2;
-	twostrong = (testlimit/10)*1;
-	simple = testlimit - onestrong - twostrong;
-
-	// now, sort the table
-	for (i = 0; i < keylen; i++) {
-                qsort(&table[i][0], n, sizeof(PTW_tableentry), &compare);
-		strongbytes[i] = 0;
-        }
-
-	sh = alloca(sizeof(sorthelper) * (n-1) * keylen);
-	if (sh == NULL) {
-		printf("could not allocate memory\n");
-		exit(-1);
-	}
-
-
-	for (i = 0; i < keylen; i++) {
-		for (j = 1; j < n; j++) {
-			sh[i][j-1].distance = table[i][0].votes - table[i][j].votes;
-			sh[i][j-1].value = table[i][j].b;
-			sh[i][j-1].keybyte = i;
+	if(!(attacks & NO_KLEIN))
+	{
+		// Try the original klein attack first
+		for (i = 0; i < keylen; i++) {
+			memset(&table[i][0], 0, sizeof(PTW_tableentry) * n);
+			for (j = 0; j < n; j++) {
+				table[i][j].b = j;
+			}
+			for (j = 0; j < state->packets_collected; j++) {
+				// fullkeybuf[0] = state->allsessions[j].iv[0];
+				memcpy(fullkeybuf, state->allsessions[j].iv, 3 * sizeof(uint8_t));
+				guesskeybytes(i+3, fullkeybuf, state->allsessions[j].keystream, guessbuf, 1);
+				table[i][guessbuf[0]].votes += state->allsessions[j].weight;
+			}
+			qsort(&table[i][0], n, sizeof(PTW_tableentry), &compare);
+			j = 0;
+			while(!validchars[i][table[i][j].b]) {
+				j++;
+			}
+			// printf("guessing i = %d, b = %d\n", i, table[0][0].b);
+			fullkeybuf[i+3] = table[i][j].b;
+		}
+		if (correct(state, &fullkeybuf[3], keylen)) {
+			memcpy(keybuf, &fullkeybuf[3], keylen * sizeof(uint8_t));
+			// printf("hit without correction\n");
+			return 1;
 		}
 	}
-	qsort(sh, (n-1)*keylen, sizeof(sorthelper), &comparesorthelper);
 
 
-	if (doComputation(state, keybuf, keylen, table, (sorthelper *) sh, strongbytes, simple)) {
-		return 1;
+	if(!(attacks & NO_PTW))
+	{
+		memcpy(table, state->table, sizeof(PTW_tableentry) * n * keylen);
+
+		onestrong = (testlimit/10)*2;
+		twostrong = (testlimit/10)*1;
+		simple = testlimit - onestrong - twostrong;
+
+		// now, sort the table
+		for (i = 0; i < keylen; i++) {
+			qsort(&table[i][0], n, sizeof(PTW_tableentry), &compare);
+			strongbytes[i] = 0;
+		}
+
+		sh = alloca(sizeof(sorthelper) * (n-1) * keylen);
+		if (sh == NULL) {
+			printf("could not allocate memory\n");
+			exit(-1);
+		}
+
+
+		for (i = 0; i < keylen; i++) {
+			for (j = 1; j < n; j++) {
+				sh[i][j-1].distance = table[i][0].votes - table[i][j].votes;
+				sh[i][j-1].value = table[i][j].b;
+				sh[i][j-1].keybyte = i;
+			}
+		}
+		qsort(sh, (n-1)*keylen, sizeof(sorthelper), &comparesorthelper);
+
+
+		if (doComputation(state, keybuf, keylen, table, (sorthelper *) sh, strongbytes, simple, bf, validchars)) {
+			return 1;
+		}
+
+		// Now one strong byte
+		getdrv(state->table, keylen, normal, ausreisser);
+		for (i = 0; i < keylen-1; i++) {
+			helper[i].keybyte = i+1;
+			helper[i].difference = normal[i+1] - ausreisser[i+1];
+		}
+		qsort(helper, keylen-1, sizeof(doublesorthelper), &comparedoublesorthelper);
+		// do not use bf-bytes as strongbytes
+		i = 0;
+		while(bf[helper[i].keybyte] == 1) {
+			i++;
+		}
+		strongbytes[helper[i].keybyte] = 1;
+		if (doComputation(state, keybuf, keylen, table, (sorthelper *) sh, strongbytes, onestrong, bf, validchars)) {
+			return 1;
+		}
+
+		// two strong bytes
+		i++;
+		while(bf[helper[i].keybyte] == 1) {
+			i++;
+		}
+		strongbytes[helper[i].keybyte] = 1;
+		if (doComputation(state, keybuf, keylen, table, (sorthelper *) sh, strongbytes, twostrong, bf, validchars)) {
+			return 1;
+		}
 	}
-
-	// Now one strong byte
-	getdrv(state->table, keylen, normal, ausreisser);
-	for (i = 0; i < keylen-1; i++) {
-		helper[i].keybyte = i+1;
-		helper[i].difference = normal[i+1] - ausreisser[i+1];
-	}
-	qsort(helper, keylen-1, sizeof(doublesorthelper), &comparedoublesorthelper);
-	strongbytes[helper[0].keybyte] = 1;
-	if (doComputation(state, keybuf, keylen, table, (sorthelper *) sh, strongbytes, onestrong)) {
-		return 1;
-	}
-
-	// two strong bytes
-	strongbytes[helper[1].keybyte] = 1;
-	if (doComputation(state, keybuf, keylen, table, (sorthelper *) sh, strongbytes, twostrong)) {
-		return 1;
-	}
-
 	return 0;
 }
 
@@ -427,8 +551,8 @@ int PTW_computeKey(PTW_attackstate * state, uint8_t * keybuf, int keylen, int te
  * iv - IV used in the session
  * keystream - recovered keystream from the session
  */
-int PTW_addsession(PTW_attackstate * state, uint8_t * iv, uint8_t * keystream) {
-	int i;
+int PTW_addsession(PTW_attackstate * state, uint8_t * iv, uint8_t * keystream, int * weight, int total) {
+	int i,j;
 	int il;
 	int ir;
 	uint8_t buf[PTW_KEYHSBYTES];
@@ -437,17 +561,31 @@ int PTW_addsession(PTW_attackstate * state, uint8_t * iv, uint8_t * keystream) {
 	il = i/8;
 	ir = 1 << (i%8);
 	if ((state->seen_iv[il] & ir) == 0) {
-		state->packets_collected++;
 		state->seen_iv[il] |= ir;
-		guesskeybytes(iv, keystream, buf, PTW_KEYHSBYTES);
-                for (i = 0; i < KEYHSBYTES; i++) {
-                	state->table[i][buf[i]].votes++;
-                }
-		if (state->sessions_collected < CONTROLSESSIONS) {
-			memcpy(state->sessions[state->sessions_collected].iv, iv, IVBYTES);
-			memcpy(state->sessions[state->sessions_collected].keystream, keystream, KSBYTES);
-			state->sessions_collected++;
+		for (j = 0; j < total; j++) {
+			state->packets_collected++;
+			guesskeybytes(IVBYTES, iv, &keystream[KSBYTES*j], buf, PTW_KEYHSBYTES);
+	                for (i = 0; i < KEYHSBYTES; i++) {
+	                	state->table[i][buf[i]].votes += weight[j];
+	                }
+			if (state->allsessions_size < state->packets_collected) {
+				state->allsessions_size = state->allsessions_size << 1;
+				state->allsessions = realloc(state->allsessions, state->allsessions_size * sizeof(PTW_session));
+				if (state->allsessions == NULL) {
+					printf("could not allocate memory\n");
+					exit(-1);
+				}
+			}
+			memcpy(state->allsessions[state->packets_collected-1].iv, iv, IVBYTES);
+			memcpy(state->allsessions[state->packets_collected-1].keystream, &keystream[KSBYTES*j], KSBYTES);
+			state->allsessions[state->packets_collected-1].weight = weight[j];
 		}
+                if ((state->sessions_collected < CONTROLSESSIONS)) {
+                        memcpy(state->sessions[state->sessions_collected].iv, iv, IVBYTES);
+                        memcpy(state->sessions[state->sessions_collected].keystream, keystream, KSBYTES);
+                        state->sessions_collected++;
+                }
+
 		return 1;
 	} else {
 		return 0;
@@ -464,12 +602,19 @@ PTW_attackstate * PTW_newattackstate() {
 	if (state == NULL) {
 		return NULL;
 	}
-	bzero(state, sizeof(PTW_attackstate));
+	memset(state, 0, sizeof(PTW_attackstate));
 	for (i = 0; i < PTW_KEYHSBYTES; i++) {
                 for (k = 0; k < n; k++) {
                         state->table[i][k].b = k;
                 }
         }
+	state->allsessions = malloc(4096 * sizeof(PTW_session));
+	state->allsessions_size = 4096;
+	if (state->allsessions == NULL) {
+		printf("could not allocate memory\n");
+		exit(-1);
+	}
+
         return state;
 }
 
@@ -477,7 +622,7 @@ PTW_attackstate * PTW_newattackstate() {
  * Free an allocated attackstate
  */
 void PTW_freeattackstate(PTW_attackstate * state) {
+	free(state->allsessions);
 	free(state);
 	return;
 }
-
