@@ -934,6 +934,12 @@ usage:
                 st_cur->eapol_size = ( h80211[z + 2] << 8 )
                                    +   h80211[z + 3] + 4;
 
+                if ((int)pkh.len - z < st_cur->eapol_size  || st_cur->eapol_size == 0)
+                {
+                	// Ignore the packet trying to crash us.
+                	continue;
+                }
+
                 memcpy( st_cur->keymic, &h80211[z + 81], 16 );
                 memcpy( st_cur->eapol, &h80211[z], st_cur->eapol_size );
                 memset( st_cur->eapol + 81, 0, 16 );
@@ -962,6 +968,12 @@ usage:
                 st_cur->eapol_size = ( h80211[z + 2] << 8 )
                                    +   h80211[z + 3] + 4;
 
+                if ((int)pkh.len - z < st_cur->eapol_size  || st_cur->eapol_size == 0)
+				{
+					// Ignore the packet trying to crash us.
+					continue;
+				}
+
                 memcpy( st_cur->keymic, &h80211[z + 81], 16 );
                 memcpy( st_cur->eapol, &h80211[z], st_cur->eapol_size );
                 memset( st_cur->eapol + 81, 0, 16 );
