@@ -1,10 +1,29 @@
-/* displays the votes of the korek attack for each keybyte */
+/*
+ *  Kstat: displays the votes of the korek attack for each keybyte
+ *
+ *  Copyright (C) 2006, 2007, 2008, 2009 Thomas d'Otreppe
+ *  Copyright (C) 2004, 2005  Christophe Devine
+ *
+ *  This program is free software; you can redistribute it and/or modify
+ *  it under the terms of the GNU General Public License as published by
+ *  the Free Software Foundation; either version 2 of the License, or
+ *  (at your option) any later version.
+ *
+ *  This program is distributed in the hope that it will be useful,
+ *  but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *  GNU General Public License for more details.
+ *
+ *  You should have received a copy of the GNU General Public License
+ *  along with this program; if not, write to the Free Software
+ *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ */
 
 #include <string.h>
 #include <stdlib.h>
 #include <stdio.h>
 
-#define SWAP(x,y) { unsigned char tmp = x; x = y; y = tmp; }
+#include "common.h"
 
 #define N_ATTACKS 17
 
